-- vim: ts=2 sw=2 ai et

local NAME = "Track-o-Matique"

ToM = LibStub ("AceAddon-3.0"):NewAddon (NAME, "AceConsole-3.0", "AceEvent-3.0")

local options = {
  type = 'group',
  handler = ToM,
  childGroups = 'tab',
  get = "GetVal", set = "SetVal",
  args = {
    options = {
      type = 'group', name = "Options", desc = NAME.." Options", order = 1,
      args = {
        automatic = {
          type = 'toggle', order = 1,
          name = "Automatic tracking",
          desc = "Automatically change tracking based on current target. " ..
            "When not using automatic tracking, use /tom in your attack macro.",
        },
        conditions = { type = 'group', order = 3,  guiInline = true,
          disabled = function() return not ToM.db.profile.automatic end,
          name = "Automatic tracking options",
          args = {
            raid = { type = 'toggle', order = 10,
              name = "Only in Raid/Party",
              desc = "Only while in a Raid or Party" },
            it = { type = 'toggle', order = 20,
              name="Improved Tracking",
              desc = "Only when talented for Improved Tracking" },
            combat = { type = 'toggle', order = 30,
              name="Only in Combat",
              desc = "Only while in combat\n(consider activating Aggressive"..
              " checking for a better tracking experience when using this)" },
            restore = { type = 'toggle', order = 80,
              name = "Restore tracking",
              desc = "Restore the tracking type to what it was before" ..
              " entering combat",
              disabled = function()
                 return not ToM.db.profile.automatic or
                        not ToM.db.profile.combat
              end
            },
          },
        },
        aggressive = { type = 'toggle', order = 100,
          name = "Aggressive checking",
          desc = "Continuously scan the current target and ensure the right"..
          " tracking option is activated. This is far more CPU intensive than"..
          " the regular target tracking, but ensures you optimize your damage.",
        },
        _ = { type = 'description', order = 110, name = " " },
        debug = {
          type = 'toggle', name = "Debugging", order = -1,
          desc = "Enable printing of debugging information. " ..
            "This is very irritating. You probably don't want this."
        }
      }
    },
    about = {
      type = 'group', name = "About", desc = "About "..NAME, order = -1,
      args = {
        __ = { type = 'header',
               name = NAME..GetAddOnMetadata (NAME,"Version"), order = 1 },
        _1 = { type = 'description',
               name = GetAddOnMetadata (NAME, "Notes"), order = 10 },
        _2 = { type = 'description',
               name = "By "..GetAddOnMetadata (NAME, "Author"), order = 20 },
        _3 = { type = 'description', name = "", order = 30 },
        _4 = { type = 'description',
               name = "Copyright(C) 2008 Johny Mattsson", order = 40 },
        _5 = { type = 'description', name = "", order = 50 },
        _6 = { type = 'description',
               name = "Build "..GetAddOnMetadata (NAME,"X-Build"):match ("%d+"),
               order = 60 }
      }
    }
  }
}

-- Default profile options
local defaults = {
  profile = {
    automatic = true,
    raid = false,
    it = true,
    combat = false,
    restore = false,
    aggressive = false,
    debug = false,
  }
}


function ToM:GetVal (info)
  return self.db.profile[info[#info]]
end

function ToM:SetVal (info, v)
  self.db.profile[info[#info]] = v
  self:CheckAggressive ()
end

-- Avoid having to iterate over GetTrackingInfo to pull out the active tracking,
-- use lookup tables based on the current tracking texture instead
ToM.type2texture = {
  Beast = "Interface\\Icons\\Ability_Tracking",
  Demon = "Interface\\Icons\\Spell_Shadow_SummonFelHunter",
  Dragonkin = "Interface\\Icons\\INV_Misc_Head_Dragon_01",
  Elemental = "Interface\\Icons\\Spell_Frost_SummonWaterElemental",
  Giant = "Interface\\Icons\\Ability_Racial_Avatar",
  Humanoid = "Interface\\Icons\\Spell_Holy_PrayerOfHealing",
  Undead = "Interface\\Icons\\Spell_Shadow_DarkSummoning"
}
-- Reverse look up table
ToM.texture2type = {}

-- Helper functions
local function dprint (msg)
  if ToM.db.profile.debug then
    ToM:Print (msg)
  end
end

local function has_improved_tracking ()
  return select (5, GetTalentInfo (3, 1)) > 0
end


function ToM:AutoTrack ()
  local dead = UnitIsDead ("target")
  local attackable = UnitCanAttack ("player", "target")
  local targetType = UnitCreatureType ("target")
  if dead or not attackable or not targetType then
    -- No point in updating tracking
    return
  end

  local currentTexture = GetTrackingTexture ()
  if self.type2texture[targetType] ~= currentTexture then
    local index = self.type2index[targetType]
    if (index) then
      dprint("Activating "..self.type2name[targetType])
      SetTracking (index)
    else
      dprint ("Target is untrackable! ("..targetType..")")
    end
    return 1
  end
end

function ToM:OnUpdate (elapsed)
  -- If we've recently attempted a tracking switch, wait for the hold-off
  -- before we try again
  if self.hold_off then
    self.hold_off = self.hold_off - elapsed
    if (self.hold_off <= 0) then
      self.hold_off = nil
    end
    return
  end

  if self:MaybeUpdateTracking () then
    self.hold_off = 0.6 -- This seems to be a pretty good value
  else
    -- Probably want to avoid at going full blast - 10Hz should be plenty!
    self.hold_off = 0.1 -- Might want to make this configurable
  end
end

function ToM:CheckAggressive ()
  if self.db.profile.aggressive then
    dprint ("aggressive checking is ON")
    self.frame:SetScript ("OnUpdate", function(_, t) ToM:OnUpdate (t) end)
  else
    dprint ("aggressive checking is OFF")
    self.frame:SetScript ("OnUpdate", nil)
  end
end

function ToM:UpdateLookupTables ()
  local count = 0
  self.type2index = {}
  self.type2name = {}
  for i = 1, GetNumTrackingTypes () do
    local name, texture, active, category = GetTrackingInfo (i)
    local typ = self.texture2type[texture]
    if typ then
      -- Unit type we can and want to track
      self.type2index[typ] = i
      self.type2name[typ] = name
      count = count + 1
    end
  end
  dprint ("Now have "..count.." tracking types enabled")
end

-- Verify all the conditional options against current conditions, and if they're
-- met, chain through to AutoTrack()
function ToM:MaybeUpdateTracking ()
  local db = self.db.profile
  if not db.automatic then return end
  if db.raid and (GetNumPartyMembers () == 0 and GetNumRaidMembers () == 0) then
    return
  end
  if db.it and not has_improved_tracking () then return end
  if db.combat and not UnitAffectingCombat ("player") then return end
  return self:AutoTrack ()
end

-- GetTrackingInfo doesn't work until the character info is loaded
function ToM:PLAYER_ENTERING_WORLD ()
  self:UpdateLookupTables ()
end
-- Might have learned a new tracking type
function ToM:LEARNED_SPELL_IN_TAB (tabId)
  self:UpdateLookupTables ()
end
-- Update tracking, maybe
function ToM:PLAYER_TARGET_CHANGED ()
  self:MaybeUpdateTracking ()
end

-- Store current tracking type for possible later restoration
function ToM:PLAYER_REGEN_DISABLED ()
  self.ooc_tracking = self.texture2type[GetTrackingTexture ()]
  self:MaybeUpdateTracking ()
end

function ToM:PLAYER_REGEN_ENABLED ()
  local db = self.db.profile
  if db.automatic and db.combat and db.restore and self.ooc_tracking then
    -- Should tie this in with the Aggressive checking probably - FIXME
    SetTracking (self.type2index[self.ooc_tracking])
    self.ooc_tracking = nil
  end
end

function ToM:OnInitialize ()
  local _,playerClass = UnitClass("player")
  if (playerClass ~= "HUNTER") then
    self.noload = true
    return
  end

  for typ, tex in pairs(self.type2texture) do
    self.texture2type[tex] = typ
  end

  LibStub ("AceConfigRegistry-3.0"):RegisterOptionsTable (NAME, options)
  local ACD = LibStub ("AceConfigDialog-3.0")
  ACD:AddToBlizOptions (NAME, GetAddOnMetadata (NAME, "Title"))
  --ACD:AddToBlizOptions (NAME, nil, nil, "options")
  --ACD:AddToBlizOptions (NAME, "About", NAME, "about")
  -- Bring up GUI config screen
  self:RegisterChatCommand ("track-o-matique", function ()
    InterfaceOptionsFrame_OpenToCategory (ACD.BlizOptions[NAME].frame) end)

  -- Run auto-track command
  self:RegisterChatCommand ("tom", function() ToM:AutoTrack () end)

  -- Profile db
  self.db = LibStub ("AceDB-3.0"):New ("Track_o_MatiqueDB", defaults)
end

function ToM:OnEnable ()
  if self.noload then
    self:Print("not loaded - this is a hunter-only add-on")
    return
  end
  self.frame = CreateFrame ("Frame", nil)
  self:CheckAggressive ()
  self:RegisterEvent("LEARNED_SPELL_IN_TAB")
  self:RegisterEvent("PLAYER_ENTERING_WORLD")
  self:RegisterEvent("PLAYER_TARGET_CHANGED")
  self:RegisterEvent("PLAYER_REGEN_DISABLED")
  self:RegisterEvent("PLAYER_REGEN_ENABLED")
  if self.db.profile.automatic then
    self:Print ("Automatic tracking now ON")
  else
    self:Print ("Automatic tracking now OFF (use /tom for manual tracking)")
  end
end

